"""Functions for applying dynamic real-valued shifts to dask arrays using Lagrange interpolation

Use make_dynamic_shift_lagrange_dask to create a Lagrange interpolator for dask arrays.
"""

from __future__ import annotations

from typing import Callable

import dask
import dask.array as da
import numpy as np

from lisainstrument.dynamic_delay_numpy import (
    DynShiftBC,
    RegularInterpLagrange,
    RegularInterpLinear,
    RegularInterpMethod,
)
from lisainstrument.fir_filters_dask import DaskArray1D, make_dask_array_1d
from lisainstrument.fir_filters_numpy import (
    EdgeHandling,
    NumpyArray1D,
    make_numpy_array_1d,
)


class DynamicShiftDask:
    """Interpolate to locations given by a non-const shift.

    This allows to interpolate samples in a dask array to locations specified
    by a shift given by another dask array of same size. The shift is specified in
    units of the array index, i.e. there is no separate coordinate array.
    A negative shift refers to values left of a given sample, positive shifts
    to values on the right.

    The boundary treatment can be specified for each boundary in terms of
    DynShiftBC enums.

    The interpolation method is not fixed but provided via an interpolator
    instance implementing the RegularInterpMethod protocol.

    For technical reasons, the shift values have to be within some bound that
    has to be provided.
    """

    def __init__(
        self,
        min_delay: float,
        max_delay: float,
        left_bound: EdgeHandling,
        right_bound: EdgeHandling,
        interp: RegularInterpMethod,
    ):
        """Set up interpolation parameters

        Arguments:
            min_delay: Assume that any shift < -min_delay
            max_delay: Assume that any shift > -max_delay
            left_bound: Treatment of left boundary
            right_bound: Treatment of right boundary
            interp: implementation of an interpolation method
        """

        if min_delay > max_delay:
            msg = f"dynamic_shift_dask: invalid delay range {min_delay=}, {max_delay=}"
            raise ValueError(msg)

        self._min_delay = int(np.floor(min_delay))
        self._max_delay = int(np.ceil(max_delay))

        self._left_bound = left_bound
        self._right_bound = right_bound
        self._interp_np = interp

    def _op_interp(
        self, samp: np.ndarray, loc: np.ndarray, ofs: np.ndarray
    ) -> np.ndarray:
        """helper method ensuring correct array dimensionalities"""
        return self._interp_np(
            make_numpy_array_1d(samp),
            make_numpy_array_1d(loc),
            make_numpy_array_1d(ofs),
        )

    @property
    def margin_left(self) -> int:
        """Left margin size.

        If positive, specifies how many samples on the left have to be added by boundary conditions.
        If negative, specifies how many samples on the left are unused.
        """
        return self._interp_np.margin_left + self._max_delay

    @property
    def margin_right(self) -> int:
        """Right margin size.

        If positive, specifies how many samples on the right have to be added by boundary conditions.
        If negative, specifies how many samples on the right are unused.
        """
        return self._interp_np.margin_right - self._min_delay

    def __call__(self, samples: da.Array, shift: da.Array) -> DaskArray1D:
        """Apply shift given by dask array to samples in another dask array.

        The shift and sample arrays need to have the same size, and each shift provides
        the interpolation location relative to the sample with the same index.
        Shifts are floating point values.A shift of +1 refers to the sample on the right,
        -1 the sample on the left, etc. All arrays have to be 1D.

        Arguments:
            samples: 1D dask array with data samples
            shift: 1D dask array with shifts

        Returns:
            Dask array with interpolated samples
        """
        npad_left = 0
        if self.margin_left > 0:
            match self._left_bound:
                case DynShiftBC.ZEROPAD:
                    npad_left = self.margin_left
                    samples = da.concatenate([da.zeros(npad_left), samples])
                case DynShiftBC.FLAT:
                    npad_left = self.margin_left
                    samples = da.concatenate([da.ones(npad_left) * samples[0], samples])
                case _:
                    msg = (
                        f"DynamicShiftDask: left edge handling {self._left_bound.name} not "
                        f"possible for given max delay {self._max_delay}."
                    )
                    raise RuntimeError(msg)

        if self.margin_right > 0:
            match self._right_bound:
                case DynShiftBC.ZEROPAD:
                    samples = da.concatenate([samples, da.zeros(self.margin_right)])
                case DynShiftBC.FLAT:
                    samples = da.concatenate(
                        [samples, da.ones(self.margin_right) * samples[-1]]
                    )
                case _:
                    msg = (
                        f"DynamicShiftDask: right edge handling {self._right_bound.name} not "
                        f"possible for given min delay {self._min_delay=}."
                    )
                    raise RuntimeError(msg)

        chunks = shift.to_delayed()
        delayed_op = dask.delayed(self._op_interp)
        results = []
        pos = 0
        for chunk, chunk_shape in zip(chunks, shift.chunks[0], strict=True):
            n_size = npad_left + chunk_shape + self.margin_right
            n_first = pos + npad_left - self.margin_left
            samples_needed = samples[n_first : n_first + n_size]

            loc = -chunk
            offsets = self.margin_left + da.arange(chunk_shape, chunks=chunk_shape)

            samples_shifted = delayed_op(samples_needed, loc, offsets)
            delayed_chunk = da.from_delayed(
                samples_shifted, (chunk_shape,), samples.dtype
            )
            results.append(delayed_chunk)
            pos += chunk_shape
        return make_dask_array_1d(da.concatenate(results, axis=0))


def make_dynamic_shift_linear_dask(
    min_delay: float,
    max_delay: float,
    left_bound: EdgeHandling,
    right_bound: EdgeHandling,
) -> DynamicShiftDask:
    """Set up DynamicShiftDask instance with linear interpolation method.

    Arguments:
        min_delay: Assume that any shift < -min_delay
        max_delay: Assume that any shift > -max_delay
        left_bound: Treatment of left boundary
        right_bound: Treatment of right boundary

    Returns:
        Interpolation function of type DynamicShiftDask
    """
    interp = RegularInterpLinear()
    return DynamicShiftDask(min_delay, max_delay, left_bound, right_bound, interp)


def make_dynamic_shift_lagrange_dask(
    length: int,
    min_delay: float,
    max_delay: float,
    left_bound: EdgeHandling,
    right_bound: EdgeHandling,
) -> DynamicShiftDask:
    """Set up DynamicShiftDask instance with Lagrange interpolation method.

    Arguments:
        length: Number of lagrange polynomials (of order length - 1)
        min_delay: Assume that any shift < -min_delay
        max_delay: Assume that any shift > -max_delay
        left_bound: Treatment of left boundary
        right_bound: Treatment of right boundary

    Returns:
        Interpolation function of type DynamicShiftDask
    """

    interp = RegularInterpLagrange(length)
    return DynamicShiftDask(min_delay, max_delay, left_bound, right_bound, interp)


def numpyfy_dask_bivariate(
    dafunc: Callable[[da.Array, da.Array], DaskArray1D], chunks: int
) -> Callable[[np.ndarray, np.ndarray], NumpyArray1D]:
    """Convert function operating on dask arrays to work with numpy arrays.

    The dask function will be called with two chunked dask arrays created from
    the numpy arrays, and then evaluate the result.

    Arguments:
        dafunc: function operating on two dask arrays returning single 1D dask array
        chunks: chunk size to be used internally
    Returns:
        function operating on numpy arrays
    """

    def func(x, y):
        dx = da.from_array(x, chunks=chunks)
        dy = da.from_array(y, chunks=chunks)
        r = dafunc(make_dask_array_1d(dx), make_dask_array_1d(dy)).compute()
        return make_numpy_array_1d(r)

    return func
